/* copyright (c)2012..2013 Oskar Schirmer <oskar@scara.com> */
#include "arithlog.h"
#include "chanend.h"
#include "ethernet.h"
#include "reception.h"
#include "resource.h"
#include "print.h"
#include "uart.h"

struct mii_rx {
  resource clkblk;
  resource port_clk;
  resource port_er;
  resource port_d;
  resource port_dv;
  resource port_led;
};

struct mii_tx {
  resource clkblk;
  resource port_clk;
  resource port_en;
  resource port_d;
  resource port_led;
};

struct smi {
  resource clkblk;
  resource port_mdio;
  resource port_mdc;
};

const static resource clkblk_mii_ref = resource_clkblk(0);

const static struct mii_rx if_mii_rx = {
  resource_clkblk(1),
  resource_port(1,12),
  resource_port(1,14),
  resource_port(4,4),
  resource_port(1,13),
  resource_port(1,10)
};

const static struct mii_tx if_mii_tx = {
  resource_clkblk(2),
  resource_port(1,9),
  resource_port(1,11),
  resource_port(4,5),
  resource_port(1,8)
};

const static struct smi if_smi = {
  resource_clkblk(5),
  resource_port(8,3),
  resource_port(1,15)
};

static void smi_init(const struct smi *smi)
{
  resource t;
  time reset;
  clkblk_off(smi->clkblk);
  clkblk_on(smi->clkblk);
  clkblk_reference(smi->clkblk, resource_clockref, (1 * time_per_usec) / 2);
  clkblk_start(smi->clkblk);
  port_off(smi->port_mdc);
  port_off(smi->port_mdio);
  port_on(smi->port_mdc);
  port_on(smi->port_mdio);
  port_config_port(smi->port_mdc, port_port_dataport);
  port_reference(smi->port_mdc, smi->clkblk);
  port_config_ready(smi->port_mdc, port_ready_noready);
  port_clrbuf(smi->port_mdc);
  port_config_port(smi->port_mdio, port_port_dataport);
  port_reference(smi->port_mdio, smi->clkblk);
  port_config_ready(smi->port_mdio, port_ready_noready);
  port_clrbuf(smi->port_mdio);
  port_config_sdelay(smi->port_mdio, 1);
  port_write(smi->port_mdc, 0);
  port_write(smi->port_mdio, 0x7f);
  port_write(smi->port_mdio, 0x00);
  t = timer_book();
  timer_wait(t, 110 * time_per_usec);
  port_write(smi->port_mdio, 0x7f);
  timer_wait(t, 110 * time_per_usec);
  timer_free(t);
}

static void smi_bit_shift_out(const struct smi *smi,
    unsigned data, int count)
{
  while (--count >= 0) {
    unsigned bit;
    bit = (data & (1 << count)) ? 0xff : 0x7f;
    port_write(smi->port_mdio, bit);
    port_write(smi->port_mdc, 0);
    port_write(smi->port_mdio, bit);
    port_write(smi->port_mdc, 1);
  }
}

static unsigned smi_bit_shift_in(const struct smi *smi,
    int count)
{
  unsigned data = 0;
  while (--count >= 0) {
    port_write(smi->port_mdc, 0);
    port_read(smi->port_mdio);
    port_write(smi->port_mdc, 1);
    data = (data << 1) | (port_read(smi->port_mdio) >> 7);
  }
  return data;
}

static unsigned smi_read(const struct smi *smi,
    unsigned addr, unsigned reg)
{
  unsigned data;
  port_write(smi->port_mdc, 0);
  port_write(smi->port_mdc, 0);
  port_write(smi->port_mdio, 0x7f);
  smi_bit_shift_out(smi, 0xffffffff, 32);
  smi_bit_shift_out(smi, 0x6, 4);
  smi_bit_shift_out(smi, addr, 5);
  smi_bit_shift_out(smi, reg, 5);
  port_write(smi->port_mdc, 0);
  port_read(smi->port_mdio);
  smi_bit_shift_in(smi, 2);
  data = smi_bit_shift_in(smi, 16);
  port_write(smi->port_mdc, 0);
  return data;
}

static void smi_write(const struct smi *smi,
    unsigned addr, unsigned reg, unsigned data)
{
  port_write(smi->port_mdc, 0);
  port_write(smi->port_mdc, 0);
  port_write(smi->port_mdio, 0x7f);
  smi_bit_shift_out(smi, 0xffffffff, 32);
  smi_bit_shift_out(smi, 0x5, 4);
  smi_bit_shift_out(smi, addr, 5);
  smi_bit_shift_out(smi, reg, 5);
  smi_bit_shift_out(smi, 2, 2);
  smi_bit_shift_out(smi, data, 16);
  port_write(smi->port_mdc, 0);
  port_write(smi->port_mdio, 0x7f);
}

static void mii_rx_init(const struct mii_rx *rx, const resource ref)
{
  port_on(rx->port_clk);
  port_config_cond(rx->port_clk, port_cond_none, 0);
  port_read(rx->port_clk);
  port_on(rx->port_d);
  port_config_buffers(rx->port_d, 32);
  port_on(rx->port_dv);
  port_on(rx->port_er);
  port_reference(rx->port_clk, ref);
  port_reference(rx->port_d, ref);
  port_reference(rx->port_dv, ref);
  port_config_paddelay(rx->port_clk, 0);
  port_config_ready(rx->port_d, port_ready_strobed);
  port_config_slave(rx->port_d, 1);
  clkblk_off(rx->clkblk);
  clkblk_on(rx->clkblk);
  clkblk_reference(rx->clkblk, rx->port_clk, 0);
  clkblk_ready(rx->clkblk, rx->port_dv);
  port_reference(rx->port_d, rx->clkblk);
  port_reference(rx->port_dv, rx->clkblk);
  clkblk_config_risedelay(rx->clkblk, 0);
  clkblk_start(rx->clkblk);
  port_clrbuf(rx->port_d);
}

static void mii_tx_init(const struct mii_tx *tx, const resource ref)
{
  port_on(tx->port_clk);
  port_on(tx->port_d);
  port_config_buffers(tx->port_d, 32);
  port_on(tx->port_en);
  port_reference(tx->port_clk, ref);
  port_reference(tx->port_d, ref);
  port_reference(tx->port_en, ref);
  port_config_paddelay(tx->port_clk, 0);
  port_write(tx->port_d, 0);
  port_write(tx->port_en, 0);
  port_sync(tx->port_d);
  port_sync(tx->port_en);
  port_config_ready(tx->port_d, port_ready_strobed);
  port_config_slave(tx->port_d, 0);
  port_clrbuf(tx->port_d);
  port_ready(tx->port_en, tx->port_d);
  port_config_port(tx->port_en, port_port_readyport);
  clkblk_on(tx->clkblk);
  clkblk_reference(tx->clkblk, tx->port_clk, 0);
  port_reference(tx->port_d, tx->clkblk);
  port_reference(tx->port_en, tx->clkblk);
  clkblk_config_falldelay(tx->clkblk, 7);
  clkblk_start(tx->clkblk);
  port_clrbuf(tx->port_d);
}

#define PHY_ADDRESS 0x1f
#define MII_BMCR_REG 0
#define MII_BMCR_ANEGRES     (1 << 9)
#define MII_BMCR_ANEGENA     (1 << 12)
#define MII_PID1_REG 2
#define MII_PID2_REG 3
#define phy_is_smsc8700(id)  ((id & 0xfffffff0) == 0x0007c0c0)
#define MII_ADVR_REG 4
#define MII_ADVR_MBIT10FULL  (1 << 6)
#define MII_ADVR_MBIT100FULL (1 << 8)

static int phy_init(const struct smi *smi, int mbit100)
{
  unsigned r;
  r = (smi_read(smi, PHY_ADDRESS, MII_PID1_REG) << 16) |
    smi_read(smi, PHY_ADDRESS, MII_PID2_REG);
  if (!phy_is_smsc8700(r))
    return 1;
  r = smi_read(smi, PHY_ADDRESS, MII_ADVR_REG);
  r &= 0xfc1f;
  if (mbit100)
    r |= MII_ADVR_MBIT100FULL;
  else
    r |= MII_ADVR_MBIT10FULL;
  smi_write(smi, PHY_ADDRESS, MII_ADVR_REG, r);
  if (smi_read(smi, PHY_ADDRESS, MII_ADVR_REG) != r)
    return 1;
  r = smi_read(smi, PHY_ADDRESS, MII_BMCR_REG);
  r &= ~MII_BMCR_ANEGENA;
  smi_write(smi, PHY_ADDRESS, MII_BMCR_REG, r);
  r |= MII_BMCR_ANEGENA;
  smi_write(smi, PHY_ADDRESS, MII_BMCR_REG, r);
  r |= MII_BMCR_ANEGRES;
  smi_write(smi, PHY_ADDRESS, MII_BMCR_REG, r);
  return 0;
}

//#define DEBUG_ETHRX

#define MAX_FRAMELEN 1522

static void eth_rx(const resource arp, const resource ipv4,
    const resource dv, const resource d, const resource led)
{
  static unsigned char data[MAX_FRAMELEN+3];
#ifdef DEBUG_ETHRX
  resource log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
#endif
  port_on(led);
  port_config_cond(dv, port_cond_eq, 0);
  port_read(dv);
  do {
    const unsigned poly = 0xedb88320;
    unsigned crc = 0x9226f562;
    unsigned v, e;
    unsigned n, i = 0;
    unsigned *word = (unsigned *)&data[0];
    event_book();
    event_vector(d, eth_rx_data);
    event_vector(dv, eth_rx_ndv);
    event_enable(d);
    event_enable(dv);
    port_config_cond(dv, port_cond_eq, 1);
    port_read(dv);
    port_config_cond(d, port_cond_eq, 0xd);
    port_read(d);
    port_write(led, 1);
    do {
      port_config_cond(dv, port_cond_eq, 0);
      event_await();
      event_avoid_dce(eth_rx_data);
      event_avoid_dce(eth_rx_ndv);
event_label(eth_rx_data):
      v = port_read(d);
      if (i < (MAX_FRAMELEN+3)/4) {
        word[i++] = v;
        crc32_word(crc, v, poly);
      }
    } while (1);
event_label(eth_rx_ndv):
    port_read(dv);
    e = port_remaining(d);
    v = port_read(d);
    if (!(e & 7)) {
      v = v >> (32 - e);
      e = e >> 3;
      if (i < (MAX_FRAMELEN+3)/4) {
        word[i] = v;
        switch (e) {
          case 3:
            v = crc32_byte(crc, v, poly);
          case 2:
            v = crc32_byte(crc, v, poly);
          case 1:
            v = crc32_byte(crc, v, poly);
          case 0:
            break;
        }
      }
      i = (i << 2) + e;
      if (!((~crc) || (i < 64))) {
        resource c = resource_nil;
#ifdef DEBUG_ETHRX
        printstr(log, "eth rx: ");
        printhex(log, word[0], 8); printchar(log, ' ');
        printhex(log, word[1], 8); printchar(log, ' ');
        printhex(log, word[2], 8); printchar(log, ' ');
        printhex(log, word[3], 8); printchar(log, ' ');
        printhex(log, word[4], 8); printchar(log, ' ');
        println(log);
#endif
        switch (((unsigned short *)&data[0])[6]) {
          case 0x0608:
            c = arp;
            break;
          case 0x0008:
            c = ipv4;
            break;
        }
        if (c != resource_nil) {
          chanend_write_lsb(c, &data[14], i - 14 - 4);
          chanend_all_sent(c);
        }
      }
    }
    port_write(led, 0);
  } while (1);
#ifdef DEBUG_ETHRX
  chanend_free(log);
#endif
}

//#define DEBUG_ETHTX

static void eth_tx(const resource tx, const resource d, const resource led)
{
  const unsigned poly = 0xedb88320;
  static unsigned char data[MAX_FRAMELEN+3];
#ifdef DEBUG_ETHTX
  resource log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
#endif
  port_on(led);
  do {
    unsigned i, n, t;
    unsigned crc = 0x9226f562;
    unsigned *word = (unsigned *)&data[0];
    i = chanend_read_lsb(tx, &data[0], MAX_FRAMELEN - 4);
    for (n = 0; n < i; n++) {
      crc32_byte(crc, data[n], poly);
    }
    i += chanend_read_strip(tx);
    chanend_all_received(tx);
    while (i < 60) {
      crc32_byte(crc, 0, poly);
      data[i++] = 0;
    }
    crc32_word(crc, 0, poly);
    port_write(led, 1);
#ifdef DEBUG_ETHTX
    printstr(log, "eth tx: ");
    printhex(log, i, 4); printchar(log, ' ');
    printhex(log, word[0], 8); printchar(log, '/');
    printhex(log, word[1], 8); printchar(log, '/');
    printhex(log, word[2], 8); printchar(log, '/');
    printhex(log, word[3], 8); println(log); printchar(log, '\t');
    printhex(log, word[4], 8); printchar(log, '/');
    printhex(log, word[5], 8); printchar(log, '/');
    printhex(log, word[6], 8); printchar(log, '/');
    printhex(log, word[7], 8); println(log); printchar(log, '\t');
    printhex(log, word[8], 8); printchar(log, '/');
    printhex(log, word[9], 8); printchar(log, '/');
    printhex(log, word[10], 8); printchar(log, '/');
    printhex(log, word[11], 8); println(log); printchar(log, '\t');
    printhex(log, word[12], 8); printchar(log, '/');
    printhex(log, word[13], 8); printchar(log, '/');
    printhex(log, word[14], 8); printchar(log, ' ');
    printhex(log, ~crc, 8); println(log);
#endif
    port_write(d, 0x55555555);
    port_write(d, 0x55555555);
    port_write(d, 0xd5555555);
    t = 0;
    while (t < (i >> 2)) {
      port_write(d, word[t++]);
    }
    t = word[t];
    switch (i & 3) {
      case 3:
        port_write_partial(d, t, 24);
        break;
      case 2:
        port_write_partial(d, t, 16);
        break;
      case 1:
        port_write_partial(d, t, 8);
        break;
    }
    port_write(d, ~crc);
    port_write(led, 0);
  } while (1);
#ifdef DEBUG_ETHTX
  chanend_free(log);
#endif
}

//#define DEBUG_ETHARP

#define arp_rq_setownip 0
#define arp_rq_retrieve 1
#define arp_rq_getownip 2

static void eth_arp(const resource rx, const resource rq)
{
  const static unsigned char request_hdr[8] =
      {0x00, 0x01, 0x08, 0x00, 0x06, 0x04, 0x00, 0x01};
  static struct {
    unsigned char ip4[4];
    unsigned char mac[6];
  } table[16] = {
    {{0, 0, 0, 0}, {0x00, 0x22, 0x97, 0x00, 0x11, 0xe4}},
  };
  static unsigned char data[MAX_FRAMELEN - 14 - 4];
  unsigned x = 0;
  resource tx = chanend_book();
#ifdef DEBUG_ETHARP
  resource log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
#endif
  chanend_destination(tx, chanend_cr_request(tx, cr_eth_tx));
  event_book();
  event_vector(rx, eth_arp_rx);
  event_vector(rq, eth_arp_rq);
  event_enable(rx);
  event_enable(rq);
  do {
    unsigned i;
    event_await();
    event_avoid_dce(eth_arp_rx);
    event_avoid_dce(eth_arp_rq);
event_label(eth_arp_rx):
    i = chanend_read_lsb(rx, &data[0], 0);
    chanend_all_received(rx);
#ifdef DEBUG_ETHARP
    printstr(log, "eth arp: ");
    printhex(log, i, 4); printchar(log, ' ');
    printhex(log, ((unsigned *)&data[0])[0], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[1], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[2], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[3], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[4], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[5], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[6], 8); println(log);
#endif
    if ((((unsigned short *)&data[0])[0] == 0x0100) &&
        (((unsigned short *)&data[0])[1] == 0x0008) &&
        (data[4] == 6) &&
        (data[5] == 4)) {
      switch (((unsigned short *)&data[0])[3]) {
        case 0x0100:
#ifdef DEBUG_ETHARP
          printstr(log, "eth arp: REQUEST");
          println(log);
#endif
          if ((data[24] == table[0].ip4[0]) &&
              (data[25] == table[0].ip4[1]) &&
              (data[26] == table[0].ip4[2]) &&
              (data[27] == table[0].ip4[3])) {
            chanend_write_lsb(tx, &data[8], 6);
            chanend_write_lsb(tx, &table[0].mac[0], 6);
            chanend_write_byte(tx, 0x08);
            chanend_write_byte(tx, 0x06);
            ((unsigned short *)&data[0])[3] = 0x0200;
            chanend_write_lsb(tx, &data[0], 8);
            chanend_write_lsb(tx, &table[0].mac[0], 6);
            chanend_write_lsb(tx, &table[0].ip4[0], 4);
            chanend_write_lsb(tx, &data[8], 10);
            chanend_all_sent(tx);
          }
          break;
        case 0x0200:
#ifdef DEBUG_ETHARP
          printstr(log, "eth arp: RESPONSE");
          println(log);
#endif
          if (++x == 16) {
            x = 1;
          }
          for (i = 0; i < 6; i++) {
            table[x].mac[i] = data[i + 8];
          }
          for (i = 0; i < 4; i++) {
            table[x].ip4[i] = data[i + 14];
          }
          break;
      }
    }
    continue;
event_label(eth_arp_rq):
    chanend_listen(rq);
    i = chanend_read_byte(rq);
#ifdef DEBUG_ETHARP
    printstr(log, "eth arp: lookup/");
    printhex(log, i, 2);
    println(log);
#endif
    switch (i) {
      case arp_rq_setownip:
        chanend_read_lsb(rq, &data[0], 4);
        for (i = 0; i < 4; i++) {
          table[0].ip4[i] = data[i];
        }
        i = 0;
        break;
      case arp_rq_retrieve:
        chanend_read_lsb(rq, &data[0], 4);
        i = 0;
        while ((i < 16) &&
               ((table[i].ip4[0] != data[0]) ||
                (table[i].ip4[1] != data[1]) ||
                (table[i].ip4[2] != data[2]) ||
                (table[i].ip4[3] != data[3]))) {
          i++;
        }
        break;
      case arp_rq_getownip:
        chanend_write_lsb(rq, &table[0].ip4[0], 4);
        chanend_hangup(rq);
        continue;
      default:
        chanend_hangup(rq);
        continue;
    }
#ifdef DEBUG_ETHARP
    printstr(log, "eth arp: lookup IP: ");
    printhex(log, data[0], 2);
    printhex(log, data[1], 2);
    printhex(log, data[2], 2);
    printhex(log, data[3], 2);
    println(log);
#endif
    if (i < 16) {
#ifdef DEBUG_ETHARP
      printstr(log, "eth arp: lookup mac: ");
      printhex(log, table[i].mac[0], 2);
      printhex(log, table[i].mac[1], 2);
      printhex(log, table[i].mac[2], 2);
      printhex(log, table[i].mac[3], 2);
      printhex(log, table[i].mac[4], 2);
      printhex(log, table[i].mac[5], 2);
      printstr(log, " at: ");
      printhex(log, i, 2);
      println(log);
#endif
      chanend_write_lsb(rq, &table[i].mac[0], 6);
      chanend_hangup(rq);
    } else {
#ifdef DEBUG_ETHARP
      printstr(log, "eth arp: lookup fail");
      println(log);
#endif
      chanend_hangup(rq);
      for (i = 0; i < 6; i++) {
        chanend_write_byte(tx, 0xff);
      }
      chanend_write_lsb(tx, &table[0].mac[0], 6);
      chanend_write_byte(tx, 0x08);
      chanend_write_byte(tx, 0x06);
      chanend_write_lsb(tx, &request_hdr[0], 8);
      chanend_write_lsb(tx, &table[0].mac[0], 6);
      chanend_write_lsb(tx, &table[0].ip4[0], 4);
      for (i = 0; i < 6; i++) {
        chanend_write_byte(tx, 0x00);
      }
      chanend_write_lsb(tx, &data[0], 4);
      chanend_all_sent(tx);
    }
  } while (1);
  chanend_free(tx);
#ifdef DEBUG_ETHARP
  chanend_free(log);
#endif
}

//#define DEBUG_ETHIPV4

#define ipv4_deliv_local 0
#define ipv4_deliv_eth   1

static void eth_ipv4(const resource rx)
{
  static struct {
    unsigned char ip4[4];
    unsigned char mask[4];
    unsigned type;
    unsigned dest;
  } table[16] = {
    {{172, 23, 88, 2}, {255, 255, 255, 255}, ipv4_deliv_local, 0},
    {{0, 0, 0, 0}, {0, 0, 0, 0}, ipv4_deliv_eth, cr_eth_tx}
  };
  static unsigned char data[MAX_FRAMELEN - 14 - 4];
  static unsigned char my_mac[6];
  resource tx = chanend_book();
#ifdef DEBUG_ETHIPV4
  resource log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
#endif
  chanend_connect(tx, chanend_cr_request(tx, cr_eth_arp_rq));
  chanend_write_byte(tx, arp_rq_setownip);
  chanend_write_lsb(tx, &table[0].ip4[0], 4);
  chanend_read_lsb(tx, &my_mac[0], 6);
  chanend_hangup(tx);
  do {
    unsigned i, n, s, ihl;
    unsigned *word = (unsigned *)&data[0];
    i = chanend_read_lsb(rx, &data[0], 0);
    chanend_all_received(rx);
#ifdef DEBUG_ETHIPV4
    printstr(log, "eth ipv4: ");
    printhex(log, i, 4); printchar(log, ' ');
    printhex(log, ((unsigned *)&data[0])[0], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[1], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[2], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[3], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[4], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[5], 8); printchar(log, '/');
    printhex(log, ((unsigned *)&data[0])[6], 8); println(log);
#endif
    s = 0;
    n = ihl = 4 * (data[0] & 0x0f);
    while (n > 0) {
      n -= 2;
      s += *(unsigned short *)&data[n];
    }
    s = (unsigned short)~(s + (s >> 16));
    if (((data[0] & 0xf0) == 0x40) && /* version=4 */
#if 0
        ((((unsigned short *)&data[0])[3] & 0xffbf) == 0) && /* unfragmented */
#endif
        (data[8] != 0) && /* ttl */
        (s == 0) &&
        (ihl >= 20)) {
      unsigned char mac[6];
      n = 0;
      while (((data[16] & table[n].mask[0]) != table[n].ip4[0]) ||
             ((data[17] & table[n].mask[1]) != table[n].ip4[1]) ||
             ((data[18] & table[n].mask[2]) != table[n].ip4[2]) ||
             ((data[19] & table[n].mask[3]) != table[n].ip4[3])) {
        n++;
      }
#ifdef DEBUG_ETHIPV4
      printstr(log, "eth ipv4: protocol ");
      printhex(log, data[9], 2);
      printstr(log, ", ");
      printhex(log, n, 2);
      println(log);
#endif
      s = 0;
/*
fragments: accept fragment, when
1. it is first (offset=0)
2. it fits directly preceeding fragement with sameidentification
3. if not final, send up but dont close connection.
4. needs a timeout for unfinished pakets.
*/
      switch (table[n].type) {
        case ipv4_deliv_local:
          switch (data[9]) {
            case 1:
              s = cr_eth_icmp;
              break;
            case 6:
//            s = cr_eth_tcp;
              break;
            case 17:
              s = cr_eth_udp_rx;
              break;
          }
          if (s) {
#ifdef DEBUG_ETHIPV4
            printstr(log, "eth ipv4: local c=");
            printhex(log, s, 4);
            println(log);
#endif
            chanend_destination(tx, chanend_cr_request(tx, s));
            chanend_write_lsb(tx, &data[12], 8);
            chanend_write_lsb(tx, &data[ihl], i - ihl);
            chanend_all_sent(tx);
          }
          break;
        case ipv4_deliv_eth:
          chanend_connect(tx, chanend_cr_request(tx, cr_eth_arp_rq));
          chanend_write_byte(tx, arp_rq_retrieve);
          chanend_write_lsb(tx, &data[16], 4);
          s = chanend_read_lsb(tx, &mac[0], 6);
          chanend_hangup(tx);
#ifdef DEBUG_ETHIPV4
          printstr(log, "eth ipv4: remote mac: (");
          printhex(log, s, 2);
          printstr(log, ") ");
          printhex(log, mac[0], 2);
          printhex(log, mac[1], 2);
          printhex(log, mac[2], 2);
          printhex(log, mac[3], 2);
          printhex(log, mac[4], 2);
          printhex(log, mac[5], 2);
          println(log);
#endif
          if (s == 6) {
            data[8] -= 1;
            ((unsigned short *)&data[0])[5] += 1;
            chanend_destination(tx, chanend_cr_request(tx, table[n].dest));
            chanend_write_lsb(tx, &mac[0], 6);
            chanend_write_lsb(tx, &my_mac[0], 6);
            chanend_write_byte(tx, 0x08);
            chanend_write_byte(tx, 0x00);
            chanend_write_lsb(tx, &data[0], i);
            chanend_all_sent(tx);
          }
          break;
      }
    } else {
#ifdef DEBUG_ETHIPV4
      printstr(log, "eth ipv4: no good, ");
      printhex(log, data[0], 2); printchar(log, ' ');
      printhex(log, (((unsigned short *)&data[0])[3] & 0xffbf), 8);
      printchar(log, ' ');
      printhex(log, data[8], 2); printstr(log, ", s=");
      printhex(log, s, 4); printstr(log, ", ihl=");
      printhex(log, ihl, 4);
      println(log);
#endif
    }
  } while (1);
  chanend_free(tx);
#ifdef DEBUG_ETHIPV4
  chanend_free(log);
#endif
}

//#define DEBUG_ETHICMP

static void eth_icmp(const resource rx)
{
  static unsigned char data[MAX_FRAMELEN - 14 - 4 - 12];
  static unsigned char ip4h[12];
  resource tx = chanend_book();
#ifdef DEBUG_ETHICMP
  resource log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
#endif
  chanend_destination(tx, chanend_cr_request(tx, cr_eth_ipv4));
  do {
    unsigned i, n, s;
    short v;
    i = chanend_read_lsb(rx, &data[0], 0);
    chanend_all_received(rx);
    s = 0;
    n = i;
    if (n & 1) {
      data[n] = 0;
    }
    while (n > 8) {
      n = (n - 1) & ~1;
      s += *(unsigned short *)&data[n];
    }
    s = (unsigned short)~(s + (s >> 16));
#ifdef DEBUG_ETHICMP
    printstr(log, "eth icmp: chk=");
    printhex(log, s, 4); printchar(log, ' ');
    printhex(log, ((unsigned *)&data[8])[0], 8); printchar(log, ' ');
    printhex(log, ((unsigned *)&data[8])[1], 8); printchar(log, ' ');
    printhex(log, ((unsigned *)&data[8])[2], 8); printchar(log, ' ');
    printhex(log, ((unsigned *)&data[8])[3], 8);
    println(log);
#endif
    if (s == 0) {
      switch (data[8]) {
        case 8:
          ip4h[0] = 0x45;
          ip4h[1] = 0x00;
          ip4h[2] = (i + 12) >> 8;
          ip4h[3] = i + 12;
          ((unsigned *)&ip4h[0])[1] = 0;
          ip4h[8] = 0xff;
          ip4h[9] = 1;
          s = ((unsigned short *)&ip4h[0])[4];
          for (n = 0; n < 4; n++) {
            s += ((unsigned short *)&ip4h[0])[n];
            s += ((unsigned short *)&data[0])[n];
          }
          ((unsigned short *)&ip4h[0])[5] = ~(s + (s >> 16));
          chanend_write_lsb(tx, &ip4h[0], 12);
          chanend_write_lsb(tx, &data[4], 4);
          chanend_write_lsb(tx, &data[0], 4);
          data[8] = 0;
          data[9] = 0;
          s = 0;
          n = i;
          while (n > 12) {
            n = (n - 1) & ~1;
            s += *(unsigned short *)&data[n];
          }
          ((unsigned short *)&data[8])[1] = ~(s + (s >> 16));
          chanend_write_lsb(tx, &data[8], i - 8);
          chanend_all_sent(tx);
          break;
      }
    }
  } while (1);
  chanend_free(tx);
#ifdef DEBUG_ETHICMP
  chanend_free(log);
#endif
}

//#define DEBUG_ETHUDP

#define udp_rq_register 0
#define udp_rq_request  1
#define udp_rq_release  2

static void eth_udp(const resource rx, const resource rq)
{
  static unsigned char data[MAX_FRAMELEN - 14 - 4 - 12];
  static unsigned userport;
  static struct {
    unsigned short port;
    unsigned char used;
    resource dest;
  } table[16] = {
    {0, 0, 0}
  };
  resource tx = chanend_book();
#ifdef DEBUG_ETHUDP
  resource log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
#endif
  event_book();
  event_vector(rx, eth_udp_rx);
  event_vector(rq, eth_udp_rq);
  event_enable(rx);
  event_enable(rq);
  do {
    unsigned i, n;
    unsigned short port, len;
    resource dest;
    event_await();
    event_avoid_dce(eth_udp_rx);
    event_avoid_dce(eth_udp_rq);
event_label(eth_udp_rx):
    i = chanend_read_lsb(rx, &data[0], 0);
    chanend_all_received(rx);
    len = (data[12] << 8) + data[13];
#ifdef DEBUG_ETHUDP
    printstr(log, "eth udp: len="); printhex(log, i, 4);
    printchar(log, '/'); printhex(log, len, 4);
    println(log);
#endif
    if ((len >= 8) && ((len + 8) <= i)) {
      i = len + 8;
      port = (data[10] << 8) + data[11];
      n = 0;
      while ((n < 16) && !(table[n].used && table[n].port == port)) {
        n++;
      }
#ifdef DEBUG_ETHUDP
      printstr(log, "eth udp: port "); printhex(log, port, 4);
      printstr(log, " ("); printhex(log, n, 2); printstr(log, ") ");
      printhex(log, ((unsigned short *)&data[8])[0], 4); printchar(log, ' ');
      printhex(log, ((unsigned short *)&data[8])[1], 4); printchar(log, ' ');
      printhex(log, ((unsigned short *)&data[8])[2], 4); printchar(log, ' ');
      printhex(log, ((unsigned short *)&data[8])[3], 4); printchar(log, ':');
      printhex(log, ((unsigned *)&data[16])[0], 8); printchar(log, ' ');
      printhex(log, ((unsigned *)&data[16])[1], 8);
      println(log);
#endif
      if (n < 16) {
        dest = table[n].dest;
        chanend_destination(tx, dest);
        chanend_write_lsb(tx, &data[0], i);
        chanend_all_sent(tx);
      }
    }
    continue;
event_label(eth_udp_rq):
    chanend_listen(rq);
    i = chanend_read_byte(rq);
#ifdef DEBUG_ETHUDP
    printstr(log, "eth udp: RQ ");
    printhex(log, i, 2);
    println(log);
#endif
    switch (i) {
    case udp_rq_register:
      chanend_read_lsb(rq, &port, 2);
      chanend_read_lsb(rq, &dest, sizeof(dest));
      n = 0;
      while ((n < 16) && table[n].used && (port != table[n].port)) {
        n++;
      }
      if ((n < 16) && !table[n].used) {
        table[n].port = port;
        table[n].used = 1;
        table[n].dest = dest;
        chanend_write_lsb(rq, &port, 2);
      }
      break;
    case udp_rq_request:
      chanend_read_lsb(rq, &dest, sizeof(dest));
      n = 0;
      while ((n < 16) && table[n].used) {
        n++;
      }
#ifdef DEBUG_ETHUDP
      printstr(log, "eth udp: RQ dest ");
      printhex(log, dest, 8);
      printstr(log, ", n=");
      printhex(log, n, 2);
      println(log);
#endif
      if (n < 16) {
        port = 49152 | userport++;
        table[n].port = port;
        table[n].used = 1;
        table[n].dest = dest;
        chanend_write_lsb(rq, &port, 2);
      }
      break;
    case udp_rq_release:
      chanend_read_lsb(rq, &port, 2);
      n = 0;
      while ((n < 16) && !(table[n].used && (port == table[n].port))) {
        n++;
      }
      if (n < 16) {
        table[n].used = 0;
      }
      break;
    }
    chanend_hangup(rq);
  } while (1);
#ifdef DEBUG_ETHUDP
  chanend_free(log);
#endif
}

//#define DEBUG_ETHTFTP

static void eth_tftp(const resource rq)
{
  static unsigned char data[516 + 16];
  static unsigned char ip4h[20];
  resource t = timer_book();
  resource tx = chanend_book();
#ifdef DEBUG_ETHTFTP
  resource log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
#endif
  do {
    unsigned char cmd;
    unsigned short sprt, dprt = 69;
    unsigned short bnum = 0;
    unsigned i, s, n;
    chanend_listen(rq);
    cmd = chanend_read_byte(rq);
    switch (cmd) {
    case 'R':
      i = chanend_read_lsb(rq, &data[8 + 2], sizeof(data) - 8 - 2 - 7);
      i += 8 + 2;
      data[i++] = 0;
#ifdef DEBUG_ETHTFTP
      printstr(log, "eth tftp: file:");
      printstr(log, &data[8+2]);
      println(log);
#endif
      data[i++] = 'o';
      data[i++] = 'c';
      data[i++] = 't';
      data[i++] = 'e';
      data[i++] = 't';
      data[i++] = 0;
      ip4h[0] = 0x45;
      ip4h[1] = 0x00;
      ip4h[2] = (i + 20) >> 8;
      ip4h[3] = i + 20;
      ip4h[4] = bnum >> 8;
      ip4h[5] = bnum;
      ((unsigned short *)&ip4h[0])[3] = 0;
      ip4h[8] = 0xff;
      ip4h[9] = 17;
      ((unsigned short *)&ip4h[0])[5] = 0;
      chanend_connect(tx, chanend_cr_request(tx, cr_eth_arp_rq));
      chanend_write_byte(tx, arp_rq_getownip);
      chanend_read_lsb(tx, &ip4h[12], 4);
      chanend_hangup(tx);
#ifdef DEBUG_ETHTFTP
      printstr(log, "eth tftp: own ip:");
      printhex(log, ip4h[12], 2);
      printhex(log, ip4h[13], 2);
      printhex(log, ip4h[14], 2);
      printhex(log, ip4h[15], 2);
      println(log);
#endif
      ip4h[16] = 172;
      ip4h[17] = 23;
      ip4h[18] = 56;
      ip4h[19] = 13;
      s = 0;
      for (n = 0; n < 10; n++) {
        s += ((unsigned short *)&ip4h[0])[n];
      }
      ((unsigned short *)&ip4h[0])[5] = ~(s + (s >> 16));
      chanend_connect(tx, chanend_cr_request(tx, cr_eth_udp_rq));
      chanend_write_byte(tx, udp_rq_request);
      chanend_write_lsb(tx, &tx, sizeof(tx));
      chanend_read_lsb(tx, &sprt, 2);
      chanend_hangup(tx);
#ifdef DEBUG_ETHTFTP
      printstr(log, "eth tftp: src port: ");
      printhex(log, sprt, 4);
      println(log);
#endif
      data[0] = sprt >> 8;
      data[1] = sprt;
      data[2] = dprt >> 8;
      data[3] = dprt;
      data[4] = i >> 8;
      data[5] = i;
      data[6] = 0;
      data[7] = 0;
      data[8] = 0;
      data[9] = 1;
      chanend_destination(tx, chanend_cr_request(tx, cr_eth_ipv4));
      chanend_write_lsb(tx, &ip4h[0], 20);
      chanend_write_lsb(tx, &data[0], i);
      chanend_all_sent(tx);
      do {
#ifdef DEBUG_ETHTFTP
        printstr(log, "eth tftp: sent #");
        printhex(log, bnum, 8);
        println(log);
#endif
        bnum += 1;
        timer_wind(t, 3 * time_per_second);
        event_book();
        event_vector(tx, eth_tftp_rdat);
        event_vector(t, eth_tftp_rtmo);
        do {
          event_enable(tx);
          event_enable(t);
          event_await();
          event_avoid_dce(eth_tftp_rdat);
          event_avoid_dce(eth_tftp_rtmo);
event_label(eth_tftp_rdat):
          i = chanend_read_lsb(tx, &data[0], sizeof(data));
          chanend_all_received(tx);
#ifdef DEBUG_ETHTFTP
          printstr(log, "eth tftp: received ");
          printhex(log, i, 8);
          println(log);
#endif
          if (i >= 16 + 4) {
            unsigned j;
            if ((data[0] != 172) ||
                (data[1] != 23) ||
                (data[2] != 56) ||
                (data[3] != 13) ||
                (data[12] != ((i - 8) >> 8)) ||
                (data[13] != ((i - 8) & 0xff)) ||
                (data[16] != 0)) {
#ifdef DEBUG_ETHTFTP
              printstr(log, "eth tftp: wrong sender");
              println(log);
#endif
              continue;
            }
            if (data[17] == 5) {
#ifdef DEBUG_ETHTFTP
              printstr(log, "eth tftp: ERROR");
              println(log);
#endif
              break;
            }
            if (dprt == 69) {
              dprt = (data[8] << 8) + data[9];
            } else if ((data[8] != (dprt >> 8)) ||
                       (data[9] != (dprt & 0xff))) {
#ifdef DEBUG_ETHTFTP
              printstr(log, "eth tftp: wrong port");
              println(log);
#endif
              continue;
            }
            if ((data[17] != 3) ||
                (data[18] != (bnum >> 8)) ||
                (data[19] != (bnum & 0xff))) {
#ifdef DEBUG_ETHTFTP
              printstr(log, "eth tftp: wrong data");
              println(log);
#endif
              continue;
            }
            chanend_write_lsb(rq, &data[20], i - 20);
            chanend_not_all_sent(rq);
#ifdef DEBUG_ETHTFTP
            printstr(log, "eth tftp: ok");
            println(log);
#endif
            j = 8 + 4;
            ip4h[0] = 0x45;
            ip4h[1] = 0x00;
            ip4h[2] = (j + 20) >> 8;
            ip4h[3] = j + 20;
            ip4h[4] = bnum >> 8;
            ip4h[5] = bnum;
            ((unsigned short *)&ip4h[0])[3] = 0;
            ip4h[8] = 0xff;
            ip4h[9] = 17;
            ((unsigned short *)&ip4h[0])[5] = 0;
            ip4h[16] = 172;
            ip4h[17] = 23;
            ip4h[18] = 56;
            ip4h[19] = 13;
            s = 0;
            for (n = 0; n < 10; n++) {
              s += ((unsigned short *)&ip4h[0])[n];
            }
            ((unsigned short *)&ip4h[0])[5] = ~(s + (s >> 16));
            data[0] = sprt >> 8;
            data[1] = sprt;
            data[2] = dprt >> 8;
            data[3] = dprt;
            data[4] = j >> 8;
            data[5] = j;
            data[6] = 0;
            data[7] = 0;
            data[8] = 0;
            data[9] = 4;
            data[10] = bnum >> 8;
            data[11] = bnum;
            chanend_destination(tx, chanend_cr_request(tx, cr_eth_ipv4));
            chanend_write_lsb(tx, &ip4h[0], 20);
            chanend_write_lsb(tx, &data[0], j);
            chanend_all_sent(tx);
          }
        } while (0);
      } while (i == sizeof(data));
event_label(eth_tftp_rtmo):
#ifdef DEBUG_ETHTFTP
      printstr(log, "eth tftp: done");
      println(log);
#endif
      break;
    case 'W':
      break;
    }
    chanend_hangup(rq);
  } while (1);
}

void ethernet_init(void)
{
  resource c, t;

  clkblk_on(clkblk_mii_ref);
  port_on(if_mii_rx.port_led);
  port_on(if_mii_rx.port_dv);
  port_on(if_mii_rx.port_d);
  port_config_buffers(if_mii_rx.port_d, 32);
  port_on(if_mii_rx.port_er);
  port_on(if_mii_rx.port_clk);
  clkblk_on(if_mii_rx.clkblk);
  port_on(if_mii_tx.port_led);
  port_on(if_mii_tx.port_d);
  port_config_buffers(if_mii_tx.port_d, 32);
  port_on(if_mii_tx.port_en);
  port_on(if_mii_tx.port_clk);
  clkblk_on(if_mii_tx.clkblk);
  port_on(if_smi.port_mdc);
  port_on(if_smi.port_mdio);
  clkblk_on(if_smi.clkblk);

/*
  port_on(if_mii_rx.port_er);
  port_on(if_mii_rx.port_dv);
*/

  clkblk_on(clkblk_mii_ref);
  clkblk_start(clkblk_mii_ref);
  smi_init(&if_smi);
  mii_rx_init(&if_mii_rx, clkblk_mii_ref);
  mii_tx_init(&if_mii_tx, clkblk_mii_ref);
  t = timer_book();
  timer_wait(t, time_per_somesixthsecond);
  timer_free(t);
  phy_init(&if_smi, 1);

  c = chanend_book();
  chanend_cr_welcome(c, cr_eth_init);
  chanend_free(c);
}

void ethernet_rx(void)
{
  resource c, i;
  c = chanend_book();
  chanend_cr_request(c, cr_eth_init);
  chanend_cr_welcome(c, cr_eth_rx);
  chanend_destination(c, chanend_cr_request(c, cr_eth_arp_rx));
  i = chanend_book();
  chanend_destination(i, chanend_cr_request(i, cr_eth_ipv4));
  eth_rx(c, i, if_mii_rx.port_dv, if_mii_rx.port_d, if_mii_rx.port_led);
  chanend_free(c);
  chanend_free(i);
}

void ethernet_tx(void)
{
  resource c;
  c = chanend_book();
  chanend_cr_request(c, cr_eth_init);
  chanend_cr_welcome(c, cr_eth_tx);
  eth_tx(c, if_mii_tx.port_d, if_mii_tx.port_led);
  chanend_free(c);
}

void ethernet_arp(void)
{
  resource rx, rq;
  rx = chanend_book();
  chanend_cr_welcome(rx, cr_eth_arp_rx);
  rq = chanend_book();
  chanend_cr_welcome(rq, cr_eth_arp_rq);
  eth_arp(rx, rq);
  chanend_free(rx);
  chanend_free(rq);
}

void ethernet_ipv4(void)
{
  resource c;
  c = chanend_book();
  chanend_cr_welcome(c, cr_eth_ipv4);
  eth_ipv4(c);
  chanend_free(c);
}

void ethernet_icmp(void)
{
  resource c;
  c = chanend_book();
  chanend_cr_welcome(c, cr_eth_icmp);
  eth_icmp(c);
  chanend_free(c);
}

void ethernet_udp(void)
{
  resource rx, rq;
  rx = chanend_book();
  chanend_cr_welcome(rx, cr_eth_udp_rx);
  rq = chanend_book();
  chanend_cr_welcome(rq, cr_eth_udp_rq);
  eth_udp(rx, rq);
  chanend_free(rx);
  chanend_free(rq);
}

void ethernet_tftp(void)
{
  resource c;
  c = chanend_book();
  chanend_cr_welcome(c, cr_eth_tftp);
  eth_tftp(c);
  chanend_free(c);
}
