/* copyright (c)2012..2013 Oskar Schirmer <oskar@scara.com> */
#ifndef _RESOURCE_H
#define _RESOURCE_H

enum resource {
  res_port,
  res_timer,
  res_chanend,
  res_sync,
  res_thread,
  res_lock,
  res_clkblk,
  res_ps = 11,
  res_config
};

typedef unsigned resource;
#define resource_nil 0
#define resource_port(width, index) \
    (((width) << 16) | ((index) << 8) | res_port)
#define resource_chanend(node, core, index) \
    (((node) << 24) | ((core) << 16) | ((index) << 8) | res_chanend)
#define resource_clockref 1 /* is this resource_timer(0) ? */
#define resource_clkblk(index) \
    (((index) << 8) | res_clkblk)
#define resource_mask(a, node, core, index, type) \
    ((a) & (((node) ? 0xff << 24 : 0) | ((core) ? 0xff << 16 : 0) | \
            ((index) ? 0xff << 8 : 0) | ((type) ? 0xff : 0)))
#define resource_compatible(a, b, node, core, index, type) \
    (resource_mask((a), (node), (core), (index), (type)) == \
     resource_mask((b), (node), (core), (index), (type)))
#define resource_index(a) \
    ((unsigned char)((a) >> 8))

typedef unsigned time;
#define time_per_second 100000000
#define time_per_msec (time_per_second / 1000)
#define time_per_usec (time_per_msec / 1000)
#define time_per_somesixthsecond 16777215
#define time_per_sometwothirdmsec 65535


static inline resource resource_book(const enum resource t)
{
  resource r;
  __asm__ __volatile__ (
    "getr %0, %1" : "=r" (r) : "n" (t));
  return r;
}

static inline void resource_setd(const resource r, const unsigned d)
{
  __asm__ __volatile__ (
    "setd res[%0], %1" : : "r" (r), "r" (d));
}

static inline void resource_setc(const resource r, const unsigned c)
{
  __asm__ __volatile__ (
    "setc res[%0], %1" : : "r" (r), "nr" (c));
}

static inline void resource_setclk(const resource r, const resource c)
{
  __asm__ __volatile__ (
    "setclk res[%0], %1" : : "r" (r), "r" (c));
}

static inline void resource_setrdy(const resource r, const resource p)
{
  __asm__ __volatile__ (
    "setrdy res[%0], %1" : : "r" (r), "r" (p));
}

static inline void resource_settw(const resource r, const unsigned w)
{
  __asm__ __volatile__ (
    "settw res[%0], %1" : : "r" (r), "r" (w));
}

static inline unsigned resource_in(const resource r)
{
  unsigned v;
  __asm__ __volatile__ (
    "in %0, res[%1]" : "=r" (v) : "r" (r));
  return v;
}

static inline void resource_out(const resource r, const unsigned v)
{
  __asm__ __volatile__ (
    "out res[%0], %1" : : "r" (r), "r" (v));
}

static inline void resource_free(const resource r)
{
  __asm__ __volatile__ (
    "freer res[%0]" : : "r" (r));
}


static inline resource port_book(const unsigned width, const unsigned index)
{
  return resource_port(width, index);
}

static inline void port_off(const resource port)
{
  resource_setc(port, 0x0000);
}

static inline void port_on(const resource port)
{
  resource_setc(port, 0x0008);
  resource_setclk(port, resource_clkblk(0));
}

static inline void port_reference(const resource port, const resource clkblk)
{
  resource_setclk(port, clkblk);
}

static inline void port_ready(const resource port, const resource reference)
{
  resource_setrdy(port, reference);
}

enum {
  port_cond_none,
  port_cond_eq = 2,
  port_cond_neq,
  port_cond_greater,
  port_cond_less
};

static inline void port_config_cond(const resource port,
    const unsigned cmp, const unsigned val)
{
  if (cmp != port_cond_none)
    resource_setd(port, val);
  resource_setc(port, 0x0001 | (cmp << 3));
}

enum {
  port_drive_drive,
  port_drive_pulldown,
  port_drive_pullup
};

static inline void port_config_drive(const resource port, const unsigned val)
{
  resource_setc(port, 0x0003 | (val << 3));
}

static inline void port_clrbuf(const resource port)
{
  resource_setc(port, 0x0017);
}

static inline void port_config_slave(const resource port, const unsigned bool)
{
  resource_setc(port, 0x1007 | (bool << 3));
}

static inline void port_config_buffers(const resource port,
    const unsigned width)
{
  resource_setc(port, width ? 0x200f : 0x2007);
  if (width)
    resource_settw(port, width);
}

enum {
  port_ready_noready,
  port_ready_strobed,
  port_ready_handshake
};

static inline void port_config_ready(const resource port, const unsigned val)
{
  resource_setc(port, 0x3007 | (val << 3));
}

static inline void port_config_sdelay(const resource port, const unsigned bool)
{
  resource_setc(port, 0x4007 | (bool << 3));
}

enum {
  port_port_dataport,
  port_port_clockport,
  port_port_readyport
};

static inline void port_config_port(const resource port, const unsigned val)
{
  resource_setc(port, 0x5007 | (val << 3));
}

static inline void port_config_invert(const resource port, const unsigned bool)
{
  resource_setc(port, 0x6007 | (bool << 3));
}

static inline void port_config_paddelay(const resource port, const unsigned val)
{
  resource_setc(port, 0x7007 | (val << 3));
}

static inline unsigned port_read(const resource port)
{
  return resource_in(port);
}

static inline unsigned port_read_shift(const resource port, unsigned val)
{
  __asm__ __volatile__ (
    "inshr %0, res[%1]" : "+r" (val) : "r" (port));
  return val;
}

static inline void port_write(const resource port, const unsigned val)
{
  resource_out(port, val);
}

static inline unsigned port_write_shift(const resource port, unsigned val)
{
  __asm__ __volatile__ (
    "outshr res[%1], %0" : "+r" (val) : "r" (port));
  return val;
}

static inline void port_write_partial(const resource port, const unsigned val,
    const unsigned bits)
{
  __asm__ __volatile__ (
    "outpw res[%0], %1, %2" : : "r" (port), "r" (val), "n" (bits));
}

static inline void port_sync(const resource port)
{
  __asm__ __volatile__ (
    "syncr res[%0]" : : "r" (port));
}

static inline unsigned port_remaining(const resource port)
{
  unsigned r;
  __asm__ __volatile__ (
    "endin %0, res[%1]" : "=r" (r) : "r" (port));
  return r;
}

static inline void port_free(const resource port)
{
}


static inline resource timer_book(void)
{
  return resource_book(res_timer);
}

static inline void timer_watch(const resource timer)
{
  resource_setc(timer, 0x0001);
}

static inline void timer_alarm(const resource timer)
{
  resource_setc(timer, 0x0009);
}

static inline void timer_set(const resource timer, const time t)
{
  resource_setd(timer, t);
}

static inline time timer_read(const resource timer)
{
  return resource_in(timer);
}

static inline void timer_free(const resource timer)
{
  resource_free(timer);
}

static inline void timer_wind(const resource timer, const unsigned delay)
{
  unsigned wait;
  timer_watch(timer);
  wait = timer_read(timer) + delay;
  timer_alarm(timer);
  timer_set(timer, wait);
}

static inline void timer_wait(const resource timer, const unsigned delay)
{
  timer_wind(timer, delay);
  timer_read(timer);
}


static inline resource chanend_book(void)
{
  return resource_book(res_chanend);
}

static inline unsigned chanend_read_word(const resource chanend)
{
  return resource_in(chanend);
}

static inline unsigned char chanend_read_byte(const resource chanend)
{
  unsigned char data;
  __asm__ __volatile__ (
    "int %0, res[%1]" : "=r" (data) : "r" (chanend));
  return data;
}

static inline void chanend_write_word(const resource chanend,
    const unsigned data)
{
  resource_out(chanend, data);
}

static inline void chanend_write_byte(const resource chanend,
    const unsigned char data)
{
  __asm__ __volatile__ (
    "outt res[%0], %1" : : "r" (chanend), "r" (data));
}

static inline int chanend_nomore_byte(const resource chanend)
{
  int bool;
  __asm__ __volatile__ (
    "testct %0, res[%1]" : "=r" (bool) : "r" (chanend));
  return bool;
}

static inline int chanend_nomore_word(const resource chanend)
{
  int index;
  __asm__ __volatile__ (
    "testwct %0, res[%1]" : "=r" (index) : "r" (chanend));
  return index;
}

static inline void chanend_destination(const resource chanend,
    const resource remote)
{
  resource_setd(chanend, remote);
}

static inline void chanend_all_sent(const resource chanend)
{
  __asm__ __volatile__ (
    "outct res[%0], 1" : : "r" (chanend));
}

static inline void chanend_not_all_sent(const resource chanend)
{
  __asm__ __volatile__ (
    "outct res[%0], 2" : : "r" (chanend));
}

static inline void chanend_all_received(const resource chanend)
{
  __asm__ __volatile__ (
    "chkct res[%0], 1" : : "r" (chanend));
}

static inline resource chanend_listen(const resource chanend)
{
  resource remote;
  remote = chanend_read_word(chanend);
  chanend_destination(chanend, remote);
  return remote;
}

static inline void chanend_connect(const resource chanend,
    const resource remote)
{
  chanend_destination(chanend, remote);
  chanend_write_word(chanend, chanend);
}

static inline void chanend_hangup(const resource chanend)
{
  chanend_all_sent(chanend);
  chanend_all_received(chanend);
}

static inline void chanend_hangup_cancel(const resource chanend)
{
  chanend_all_sent(chanend);
  while (!chanend_nomore_byte(chanend))
    chanend_read_byte(chanend);
  chanend_all_received(chanend);
}

static inline void chanend_free(const resource chanend)
{
  resource_free(chanend);
}


static inline resource thread_book(void)
{
  return resource_book(res_thread);
}

static inline void thread_free(const resource thread)
{
  resource_free(thread);
}

extern resource thread_start(const unsigned short *pc,
    unsigned *dp,
    unsigned *sp,
    const unsigned *cp,
    resource obit);

extern void thread_start_preliminary(const resource thread,
    const unsigned short *pc,
    unsigned *sp);


static inline resource clkblk_book(unsigned index)
{
  return resource_clkblk(index);
}

static inline void clkblk_off(const resource clkblk)
{
  resource_setc(clkblk, 0x0000);
}

static inline void clkblk_on(const resource clkblk)
{
  resource_setc(clkblk, 0x0008);
}

static inline void clkblk_reference(const resource clkblk,
    const resource reference, const unsigned div)
{
  resource_setd(clkblk, div);
  resource_setclk(clkblk, reference);
}

static inline void clkblk_ready(const resource clkblk, const resource port)
{
  resource_setrdy(clkblk, port);
}

static inline void clkblk_stop(const resource clkblk)
{
  resource_setc(clkblk, 0x0007);
}

static inline void clkblk_start(const resource clkblk)
{
  resource_setc(clkblk, 0x000f);
}

static inline void clkblk_config_falldelay(const resource clkblk,
    const unsigned val)
{
  resource_setc(clkblk, 0x8007 | (val << 3));
}

static inline void clkblk_config_risedelay(const resource clkblk,
    const unsigned val)
{
  resource_setc(clkblk, 0x9007 | (val << 3));
}

static inline resource clkblk_free(const resource clkblk)
{
}


static inline unsigned ps_get(unsigned r)
{
  unsigned data;
  unsigned reg = (r << 8) | res_ps;
  __asm__ __volatile__ (
    "get %0, ps[%1]" : "=r" (data) : "r" (reg));
  return data;
}

static inline void ps_set(unsigned r, unsigned v)
{
  unsigned reg = (r << 8) | res_ps;
  __asm__ __volatile__ (
    "set ps[%0], %1" : : "r" (reg), "r" (v));
}


static inline void event_book(void)
{
  __asm__ __volatile__ (
    "clre");
}

#define event_vector(resource, label) ( { \
  __asm__ __volatile__ ( \
    "ldap r11, __event_a_"#label"; setv res[%0], r11" \
    : : "r" (resource) : "r11"); \
  } )

static inline void event_enable(const resource r)
{
  __asm__ __volatile__ (
    "eeu res[%0]" : : "r" (r));
}

static inline void event_await(void)
{
  __asm__ __volatile__ (
    "waiteu");
}

#define event_avoid_dce(label) ( { \
  int fake; \
  __asm__ __volatile__ ("" : "=r" (fake)); \
  if (fake) goto __event_1_##label; \
  __asm__ __volatile__ ("" : "=r" (fake)); \
  if (fake) goto __event_2_##label; \
  } )

#define event_label(label) \
  __event_1_##label: \
  __asm__ __volatile__ ("__event_a_"#label":"); \
  __event_2_##label


#endif
